#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

int IndexTable_2[] = {-1,1,-1,1};

#if 0
	int IndexTable_3[] = {-2,-1,2,4,-2,-1,2,4};
	int StepSizeTable_3[] = {1,2,3,4,5,6,7,8,10,11,13,15,17,19,22,25,28,31,35,39,44,49,54,60,67,74,82,91,101,112};
	#define stepsizetable(n) StepSizeTable_3[n]
	#define stepsizetablelen sizeof(StepSizeTable_3)
#elif 0
	int IndexTable_3[] = {-2,-1,1,2,-2,-1,1,2};
	int StepSizeTable_3[] = {1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,
								39,44,49,54,60,67,74,82,91,101,112};
	#define stepsizetable(n) StepSizeTable_3[n]
	#define stepsizetablelen sizeof(StepSizeTable_3)
#else
	int IndexTable_3[] = {-2,-1,1,2,-2,-1,1,2};
	#define stepsizetable(n) n
	#define stepsizetablelen 100
#endif

void raw_to_adpcm( unsigned short int*mem, long size, char algo, FILE* fo );
void adpcm_to_raw( unsigned char*mem, long size, char algo, FILE* fo );
char encode_8_3(int& predsample, int& index, int diff);
char encode_8_2(int& predsample, int& index, int diff);
int decode_8_3(int& predsample, int& index, char code);
int decode_8_2(int& predsample, int& index, char code);

int main (int argc, const char * argv[])
{
	if (argc < 3)
	{
		printf("%s : en/decode audio: adpcm332 <-> raw 16bit unsigned\n", argv[0]);
		printf("\n");
		printf("USAGE:");
		printf("raw to adpcm332X:  %s -e X file.raw\n", argv[0]);
		printf("adpcm332X to raw:  %s -d X file.raw\n", argv[0]);
		printf("\n");
		printf("adpcm-encoder:\n");
		printf("adpcm332a = explicit adpcm algorithm\n");
		printf("adpcm332b = optimizing algorithm: optimize current byte (3+3+2)\n");
		printf("adpcm332c = optimizing algorithm: optimize current byte + preview (3+3+2+3)\n");
		printf("adpcm332d = optimizing algorithm: optimize current byte + preview (3+3+2+3+3)\n");
		printf("adpcm332e = optimizing algorithm: optimize current byte + preview (3+3+2+3+3+2)\n");
		exit(-1);
	}
	
	// read input file
	const char *srcname = argv[3];
	FILE* f = fopen(srcname, "r" );
	if ( f==NULL)
	{
		printf("Error opening %s\n", srcname);
		exit(-2);
	}
	fseek(f, 0, SEEK_END);
	long flen = ftell(f);
	fseek(f,0,SEEK_SET);
	unsigned short int* fmem = (unsigned short int*)malloc(flen);
	if (fmem == NULL)
	{
		printf("Cannot get %li bytes of memory\n", flen);
		exit(-3);
	}
	fread(fmem, flen, 1, f);
	fclose(f);

	// what to do?
	bool encode = true;
	if ( strcmp(argv[1], "-d")==0 )
	{
		encode = false;
	}

	// open destfile
	char* destname = (char*)malloc( strlen(srcname) + 20 );
	if ( encode )
	{
		sprintf( destname, "%s.adpcm332", srcname );
	}
	else
	{
		sprintf( destname, "%s.raw", srcname );
	}	
	FILE* fo = fopen(destname, "w+" );
	if ( fo == NULL)
	{
		printf("Error opening %s\n", destname);
		exit(-4);
	}

	// actually do it...
	char algo = argv[2][0];
	if ( encode )
	{
		printf("Encoding %li words as ADPCM332%c\n", flen/2, algo);
		raw_to_adpcm( fmem, flen/2, algo, fo );
	}
	else
	{
		printf("Decoding %li bytes as ADPCM332%c\n", flen, algo);
		adpcm_to_raw( (unsigned char*)fmem, flen, algo, fo );
	}
	
	// destfile close
	fclose( fo );
	
	return 0;
}

unsigned short int minimum(unsigned short int*mem, long size)
{
	unsigned short int min = 0xFFFF;
	for(int i=0; i<size; i++)
	{
		if ( mem[i] < min )
		{
			min = mem[i];
		}
	}
	return min;
}

unsigned short int maximum(unsigned short int*mem, long size)
{
	unsigned short int max = 0x0000;
	for(int i=0; i<size; i++)
	{
		if ( mem[i] > max )
		{
			max = mem[i];
		}
	}
	return max;
}

void raw_to_adpcm( unsigned short int*mem, long size, char algo, FILE* fo )
{
	unsigned long errorsum = 0;
	
	// rescale to unsigned char
	unsigned short int min = minimum( mem, size );
	unsigned short int max = maximum(mem, size);
	unsigned short int dif = max-min;
	
	unsigned char* cmem = (unsigned char*)malloc(size);
	for ( int i=0; i<size; i++)
	{
		double d = mem[i];
#if 1
		// prescale to maximum aplitude:
		d -= min;
		d /= dif;
		d *= 249.0;
#else
		// just scale down:
		d /= 256;
#endif
		d = floor(d);
		if ( d<0 )
		{
			d=0;
		}
		if ( d>255 )
		{
			d=255;
		}
		cmem[i] = (unsigned char)d;
	}
	
	// cut silence at beginning
	long silthresh = 2;
	unsigned char sil = cmem[0];
	long silsize = 0;
	for(int i=1; i<size; i++)
	{
		if ( abs(sil-cmem[i]) > silthresh )
		{
			silsize = i-1;
			break;
		}
	}
	if ( silsize > 0 )
	{
		printf("Removed %li bytes of silence at beginning\n", silsize);
		size = size - silsize;
		cmem = cmem + silsize;
	}

	// cut silence at end
	sil = cmem[size-1];
	silsize = 0;
	for(int i=0; i<size; i++)
	{
		if ( abs(sil-cmem[size-1-i]) > silthresh )
		{
			silsize = i;
			break;
		}
	}
	if ( silsize > 0)
	{
		printf("Removed %li bytes of silence at end\n", silsize);
		size -= silsize;
	}
	
	// multiple of 3 + 1
	size = 3*((size-1)/3)+1;
	
	fwrite(cmem+0, 1, 1, fo );
	int predsample = cmem[0];
	int index = 0;
	int n = 1;
	while ( n < size )
	{
		if ( n >= size-3 || algo=='a')
		{
			int diff = cmem[n] - predsample;
			char c1 = encode_8_3(predsample, index, diff);
			errorsum += abs(cmem[n] - predsample);
			//fwrite(&code, 1, 1, fo );
			n++;

			diff = cmem[n] - predsample;
			char c2 = encode_8_3(predsample, index, diff);
			errorsum += abs(cmem[n] - predsample);
			//fwrite(&code, 1, 1, fo );
			n++;
			
			diff = cmem[n] - predsample;
			char c3 = encode_8_2(predsample, index, diff);
			errorsum += abs(cmem[n] - predsample);
			//fwrite(&code, 1, 1, fo );
			n++;
			
			char code = c1 | (c2<<3) | (c3<<6);
			fwrite(&code, 1, 1, fo );
		}
		else
		{
			int bestcode = 0;
			int besterr = 0xfffffff;
			int besterr2 = 0xfffffff;
			int bestps = predsample;
			int bestid = index;
			int maxcode2 = 0;
			switch (algo)
			{
				case 'c':
					maxcode2 = 8;
					break;
				case 'd':
					maxcode2 = 8*8;
					break;
				case 'e':
					maxcode2 = 8*8*4;
					break;
			}
			for(int code = 0; code<256; code++)
			{
				char c1 = code&7;
				char c2 = (code>>3) & 7;
				char c3 = (code>>6) & 3;
				
				int ps = predsample;
				int id = index;
				int nn = n;
				decode_8_3(ps, id, c1);
				int errorsum = abs(cmem[nn] - ps);
				nn++;
				decode_8_3(ps, id, c2);
				errorsum += abs(cmem[nn] - ps);
				nn++;
				decode_8_2(ps, id, c3);
				errorsum += abs(cmem[nn] - ps);
				nn++;
				
				if ( algo>='c' )
				{
					for (int code2 = 0; code2<maxcode2; code2++)
					{
						int errorsum2 = errorsum;
						int ps2 = ps;
						int id2 = id;
						int nn2 = nn;

						char c4 = code2&7;
						char c5 = (code2>>3) & 7;
						char c6 = (code2>>6) & 3;
						
						if ( algo>='c' )
						{
							decode_8_3(ps2, id2, c4);
							errorsum2 += abs(cmem[nn2] - ps2);
							nn2++;
						}
						if ( algo>='d' )
						{
							decode_8_3(ps2, id2, c5);
							errorsum2 += abs(cmem[nn2] - ps2);
							nn2++;
						}
						if ( algo>='e' )
						{
							decode_8_2(ps2, id2, c6);
							errorsum2 += abs(cmem[nn2] - ps2);
							nn2++;
						}
						
						if ( errorsum2 < besterr2 )
						{
							bestcode = code;
							besterr = errorsum;
							besterr2 = errorsum2;
							bestps = ps;
							bestid = id;
						}
					}
				}
				else
				{
					if ( errorsum < besterr )
					{
						bestcode = code;
						besterr = errorsum;
						bestps = ps;
						bestid = id;
					}
				}

			}
			char bestc1 = bestcode&7;
			char bestc2 = (bestcode>>3) & 7;
			char bestc3 = (bestcode>>6) & 3;
			predsample = bestps;
			index = bestid;
			//fwrite(&bestc1, 1, 1, fo );
			//fwrite(&bestc2, 1, 1, fo );
			//fwrite(&bestc3, 1, 1, fo );

			char code = bestc1 | (bestc2<<3) | (bestc3<<6);
			fwrite(&code, 1, 1, fo );

			n += 3;
			errorsum += besterr;
		}

	}
	
	double avgerr = errorsum / (1.0*size);
	printf("Average encoding error = %f\n", avgerr);
}

char encode_8_3(int& predsample, int& index, int diff)
{
	char code = 0;
	int step = stepsizetable(index);

	if ( diff >= 0 )
	{
		code = 0;
	}
	else
	{
		code = 4;
		diff = -diff;
	}

	int tempstep = step;
	if ( diff >= tempstep )
	{
		code |= 1;
		diff = diff - tempstep;
	}
	tempstep >>= 1;
	// changed from ">=" to ">" to reduce jittering (but dampens the sound)
	if ( diff > tempstep)
	{
		code |= 2;
	}
	
	int diffq = (step>>2);
	if ( code & 1 )
	{
		diffq += step;
	}
	if ( code & 2 )
	{
		diffq += (step>>1);
	}
	if (code & 4)
	{
		predsample -= diffq;
	}
	else
	{
		predsample += diffq;
	}
	
	if ( predsample>255)
		predsample=255;
	if (predsample<0)
		predsample=0;

	index += IndexTable_3[code];
	if ( index<0)
		index=0;
	if (index>=stepsizetablelen)
		index=stepsizetablelen-1;
	
	return code;
}

char encode_8_2(int& predsample, int& index, int diff)
{
	char code = 0;
	int step = stepsizetable(index);
	
	if ( diff >= 0 )
	{
		code = 0;
	}
	else
	{
		code = 2;
		diff = -diff;
	}
	
	int tempstep = step;
	if ( diff >= tempstep )
	{
		code |= 1;
		diff = diff - tempstep;
	}
	
	int diffq = (step>>1);
	if ( code & 1 )
	{
		diffq += step;
	}
	if (code & 2)
	{
		predsample -= diffq;
	}
	else
	{
		predsample += diffq;
	}
	
	if ( predsample>255)
		predsample=255;
	if (predsample<0)
		predsample=0;
	
	index += IndexTable_2[code];
	if ( index<0)
		index=0;
	if (index>=stepsizetablelen)
		index=stepsizetablelen-1;
	
	return code;
}

void adpcm_to_raw( unsigned char*mem, long size, char algo, FILE* fo )
{
	unsigned short int v = mem[0]*256;
	fwrite(&v, 2, 1, fo);
	
	int predsample = mem[0];
	int index = 0;
	int n = 1;
	int codenr = 0;
	char code = mem[n];
	while ( n < size )
	{
		switch( codenr )
		{
			case 0:
			case 1:
				decode_8_3(predsample, index, code);
				code >>= 3;
				codenr++;
				break;
			case 2:
				decode_8_2(predsample, index, code);
				codenr=0;
				n++;
				code = mem[n];
				break;
		}
		
		v = predsample*256;
		fwrite(&v, 2, 1, fo );
	}		
}

int decode_8_3(int& predsample, int& index, char code)
{
	code &= 7;
	int step = stepsizetable(index);

	index += IndexTable_3[code];
	if ( index<0)
		index=0;
	if (index>=stepsizetablelen)
		index=stepsizetablelen-1;
	
	int diffq = (step>>2);
	if ( code & 1 )
	{
		diffq += step;
	}
	if ( code & 2 )
	{
		diffq += (step>>1);
	}
	if (code & 4)
	{
		predsample -= diffq;
	}
	else
	{
		predsample += diffq;
	}

	if ( predsample>255)
		predsample=255;
	if (predsample<0)
		predsample=0;

	return predsample;
}

int decode_8_2(int& predsample, int& index, char code)
{
	code &= 3;
	int step = stepsizetable(index);
	
	index += IndexTable_2[code];
	if ( index<0)
		index=0;
	if (index>=stepsizetablelen)
		index=stepsizetablelen-1;
	
	int diffq = (step>>1);
	if ( code & 1 )
	{
		diffq += step;
	}
	if (code & 2)
	{
		predsample -= diffq;
	}
	else
	{
		predsample += diffq;
	}
	
	if ( predsample>255)
		predsample=255;
	if (predsample<0)
		predsample=0;
		
	return predsample;
}
