.include "tn25def.inc"

#define ADPCM_USE_PLL 0

;*****************************************************************
;
;	MACRO adpcm_soundout @0
;
; output sound to PWM
;*****************************************************************
.MACRO adpcm_soundout
;	out	OCR1A, @0
	out	OCR1B, @0
.ENDMACRO


;*****************************************************************
;
;	adpcm data
;
;*****************************************************************
.DSEG
;adpcm variables
adpcm_fcounter:.BYTE 1
adpcm_fcountmax:.BYTE 1		; frequency reducing counter (update pwm every N interrupts)
adpcm_codenr:	.BYTE 1		; code count 0/1/2 = 3/3/2 bits to decode
adpcm_index:	.BYTE 1		; current adpcm index
adpcm_code:	.BYTE 1		; current adpcm code (8 bit = 3+3+2)
adpcm_sound_lo:.BYTE 1		; lo pointer to sound data
adpcm_sound_hi:.BYTE 1		; hi pointer to sound data
adpcm_sample:	.BYTE 1		; current adpcm sample

.CSEG
;*****************************************************************
;
;	adpcm_init
;
; init PLL, DDR, and TIMER1
;*****************************************************************
adpcm_init:
	push	r16

#if ADPCM_USE_PLL
; PLL (wird nicht simuliert fr ATTiny25 !)
	ldi	r16, (1<<PLLE)			; Enable PLL 64MHz
	out	PLLCSR, r16
waitPLLlock:
	in	r16, PLLCSR			; PLL locked?
	sbrs r16, PLOCK					
	rjmp waitPLLlock
	in	r16, PLLCSR			; PLL = PWM clock source 
	ldi	r17, (1<<PCKE)
	or	r16, r17
	out	PLLCSR, r16			; use PLL
#endif

; DDRB
	in	r16, DDRB
	ori	r16, (1<<PB4) | (1<<PB1)		; output: PB1=OC1B, PB4=OC1A
	out	DDRB, r16

; TIMER1
#if ADPCM_USE_PLL
	; PLL/16 = 4MHz (async)
;	ldi r16, (1<<CTC1) | (1<<CS12) | (1<<CS10); | (1<<PWM1A) | (1<<COM1A0)
	; PLL/8 = 8MHz (async)
	ldi r16, (1<<CTC1) | (1<<CS12); | (1<<PWM1A) | (1<<COM1A0)
#else
	; CL/2 = 4MHz (sync)
;	ldi r16, (1<<CTC1) | (1<<CS11); | (1<<PWM1A) | (1<<COM1A0)
	; CL/1 = 8MHz
	ldi r16, (1<<CTC1) | (1<<CS10); | (1<<PWM1A) | (1<<COM1A0)
#endif
	out	TCCR1, r16
	ldi	r16, (1<<PWM1B) | (1<<COM1B0) | (1<<PSR1) 
	out	GTCCR, r16
	; 4MHz/200 = 20000Hz
	; 8MHz/200 = 40000Hz
;	ldi	r16, 199
	; 4MHz/250 = 16000Hz
	; 8MHz/250 = 32000Hz (*)
	ldi	r16, 249
	; 4MHz/256 = 15625Hz
	; 8MHz/256 = 31250Hz
;	ldi	r16, 255
	; 4MHz/125 = 32000Hz
	; 8MHz/125 = 64000Hz
	;ldi	r16, 124
	out	OCR1C, r16

; frequency reduction
	; 16000Hz / 4 = 4000Hz sample rate
;	ldi	r16, 3
	; 32000Hz / 8 = 4000Hz sample rate
	ldi	r16, 7
	; 32000Hz / 16 = 2000Hz sample rate
;	ldi	r16, 15
	sts	adpcm_fcountmax, r16
	sts	adpcm_fcounter, r16

	pop	r16
	ret

;*****************************************************************
;
;	adpcm_play
;
; play sound
; Z points to sound-structure
;*****************************************************************
adpcm_play:
	push	r16

	lpm	r16, Z+
	sts	adpcm_sample, r16
	adpcm_soundout r16
	lpm	r16, Z+
	sts	adpcm_code, r16
	clr	r16
	sts	adpcm_index, r16
	sts	adpcm_codenr, r16
	sts	adpcm_sound_lo, ZL
	sts	adpcm_sound_hi, ZH

	pop	r16
	;fall through...

;*****************************************************************
;
;	adpcm_continue
;
; continue to play sound = enable TIMER1 OVF interrupt
;*****************************************************************
adpcm_continue:
	push	r16
	in	r16, TIMSK
	ori	r16, (1<<TOIE1)
	out	TIMSK, r16
	sei

	pop	r16
	ret

;*****************************************************************
;
;	adpcm_stop
;
; stop to play sound = disable TIMER1 OVF interrupt
;*****************************************************************
adpcm_stop:
	push	r16
	in	r16, TIMSK
	andi	r16, 255-(1<<TOIE1)
	out	TIMSK, r16
	pop	r16
	ret

;*****************************************************************
;
;	adpcm_int
;
; service TIMER1 OVF interrupt
;*****************************************************************
adpcm_int:
	push r16
; frequency reduction (aka oversampling)
	lds	r16, adpcm_fcounter
	breq	adpcm_int_service
	dec	r16
	sts	adpcm_fcounter, r16
	pop	r16
	reti
; ok, it's time to do something
adpcm_int_service:
	lds	r16, adpcm_fcountmax
	sts	adpcm_fcounter, r16
;
	push	ZL
	push	ZH
	push	r17
	push	r18
	push	r19
; r19 = 0 (for ADC)
	clr r19
; r18 = index
	lds	r18, adpcm_index
; r17 = step = stepsizetable(index) = index
	mov	r17, r18
; branch according to current codenr
	lds	r16, adpcm_codenr
	inc	r16
	cpi	r16, 3
	brlt	adpcm_int_service_1_2
;
; decode 2 bits (last 2 bits)
adpcm_int_service_3:
	sts	adpcm_codenr, r19
; update index (r18) += indextable(code)
	lds	r16, adpcm_code
	andi	r16, 1
; 0 -> r16 = -1
; 1 -> r16 = +1
	lsl	r16
	dec	r16	; 0*2-1=-1 , 1*2-1=1
	add	r18, r16
	brpl	adpcm_decode2_idxPL
	clr	r18		; index<0 -> index=0
	rjmp	adpcm_decode2_idxFixed
adpcm_decode2_idxPL:
	cpi	r18, _adpcm_stepsizetable_len
	brlt	adpcm_decode2_idxFixed
				; index>tablelen -> index=tablelen
	ldi	r18, _adpcm_stepsizetable_len-1
adpcm_decode2_idxFixed:
	sts	adpcm_index, r18
; decode 2 bits
; r18 = diffq = 0
	clr	r18
; r16 = code
	lds	r16, adpcm_code
	lsr	r16
	brcc	adpcm_decode2_0x
adpcm_decode2_1x:
	add	r18, r17		; diffq += step
adpcm_decode2_0x:
	lsr	r17			; step /= 2
	add	r18, r17		; diffq += step
; r17 = sample
	lds	r17, adpcm_sample
	lsr	r16
	brcc	adpcm_decode2_x0
adpcm_decode2_x1:
	sub	r17, r18		; sample -= diffq
	brcc	adpcm_decode2_samFixed
	clr	r17			; sample<0 -> sample=0
	rjmp	adpcm_decode2_samFixed
adpcm_decode2_x0:
	add	r17, r18		; sample += diffq
	brcc	adpcm_decode2_samFixed
	ldi	r17, 255		; sample>255 -> sample=255
adpcm_decode2_samFixed:
	sts	adpcm_sample, r17

; fetch next code
	lds	ZL, adpcm_sound_lo
	lds	ZH, adpcm_sound_hi
	lpm	r16, Z+
	sts	adpcm_sound_lo, ZL
	sts	adpcm_sound_hi, ZH
	sts	adpcm_code, r16
; r17 = adpcm_sample (still!)
	rjmp	adpcm_int_service_done

; decode 3 bits (0..2 or 3..5)
adpcm_int_service_1_2:
	sts	adpcm_codenr, r16
; update index (r18) += indextable(code)
	lds	r16, adpcm_code
	andi	r16, 0b111
	ldi	ZL, low(_adpcm_indextable3*2);
	ldi	ZH, high(_adpcm_indextable3*2);
	add	ZL, r16
	adc	ZH, r19
	lpm	r16, Z
	add	r18, r16
	brpl	adpcm_decode3_idxPL
	clr	r18		; index<0 -> index=0
	rjmp	adpcm_decode3_idxFixed
adpcm_decode3_idxPL:
	cpi	r18, _adpcm_stepsizetable_len
	brlt	adpcm_decode3_idxFixed
				; index>tablelen -> index=tablelen
	ldi	r18, _adpcm_stepsizetable_len-1
adpcm_decode3_idxFixed:
	sts	adpcm_index, r18
; decode 3 bits
; r18 = diffq = 0
	clr	r18
; r16 = code
	lds	r16, adpcm_code
	lsr	r16
	brcc	adpcm_decode3_0xx
adpcm_decode3_1xx:
	add	r18, r17		; diffq += step
adpcm_decode3_0xx:
	lsr	r17			; step /= 2
	lsr	r16
	brcc	adpcm_decode3_x0x
adpcm_decode3_x1x:
	add	r18, r17		; diffq += step
adpcm_decode3_x0x:
	lsr	r17			; step /= 2
	add	r18, r17		; diffq += step
; r17 = sample
	lds	r17, adpcm_sample
	lsr	r16
	brcc	adpcm_decode3_xx0
adpcm_decode3_xx1:
	sub	r17, r18		; sample -= diffq
	brcc	adpcm_decode3_samFixed
	clr	r17			; sample<0 -> sample=0
	rjmp	adpcm_decode3_samFixed
adpcm_decode3_xx0:
	add	r17, r18		; sample += diffq
	brcc	adpcm_decode3_samFixed
	ldi	r17, 255		; sample>255 -> sample=255
adpcm_decode3_samFixed:
	sts	adpcm_code, r16
	sts	adpcm_sample, r17
	; fall through

adpcm_int_service_done:
	adpcm_soundout r17
;
	pop	r19
	pop	r18
	pop	r17
	pop	ZH
	pop	ZL
	pop	r16
	reti

_adpcm_indextable3:
	.include "adpcm_indextable3_v2.inc"

.equ	_adpcm_stepsizetable_len = 100
